package org.kantega.jexmec.mavenrepoclp;

/*
 * Copyright 2011 Kantega AS
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

import org.apache.maven.artifact.repository.metadata.io.xpp3.MetadataXpp3Reader;
import org.apache.maven.repository.internal.DefaultArtifactDescriptorReader;
import org.apache.maven.repository.internal.DefaultVersionRangeResolver;
import org.apache.maven.repository.internal.DefaultVersionResolver;
import org.apache.maven.repository.internal.MavenRepositorySystemSession;
import org.codehaus.plexus.util.xml.pull.XmlPullParserException;
import org.kantega.jexmec.PluginClassLoaderProvider;
import org.sonatype.aether.RepositorySystem;
import org.sonatype.aether.RepositorySystemSession;
import org.sonatype.aether.collection.CollectRequest;
import org.sonatype.aether.collection.DependencyCollectionException;
import org.sonatype.aether.connector.async.AsyncRepositoryConnectorFactory;
import org.sonatype.aether.connector.file.FileRepositoryConnectorFactory;
import org.sonatype.aether.graph.Dependency;
import org.sonatype.aether.graph.DependencyNode;
import org.sonatype.aether.impl.ArtifactDescriptorReader;
import org.sonatype.aether.impl.VersionRangeResolver;
import org.sonatype.aether.impl.VersionResolver;
import org.sonatype.aether.impl.internal.DefaultServiceLocator;
import org.sonatype.aether.metadata.Metadata;
import org.sonatype.aether.repository.LocalRepository;
import org.sonatype.aether.repository.RemoteRepository;
import org.sonatype.aether.resolution.ArtifactResolutionException;
import org.sonatype.aether.resolution.MetadataRequest;
import org.sonatype.aether.resolution.MetadataResult;
import org.sonatype.aether.spi.connector.RepositoryConnectorFactory;
import org.sonatype.aether.util.artifact.DefaultArtifact;
import org.sonatype.aether.util.graph.PreorderNodeListGenerator;
import org.sonatype.aether.util.metadata.DefaultMetadata;

import java.io.*;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.*;

import static java.util.Collections.singletonList;

/**
 * Provides classes by downloading a Maven jar artifact transitively and adding all files to a URL classloader.
 */
public class MavenPluginClassLoaderProvider implements PluginClassLoaderProvider {

    private final File localRepository;
    private final List<String> startArtifacts = new ArrayList<String>();
    private final Map<String,ClassLoader> classLoaders = new LinkedHashMap<String, ClassLoader>();
    private ClassLoader parentClassLoader;
    private Registry registry;
    private final RemoteRepository remoteRepository;
    private final RemoteRepository[] additionalRemoteRepositories;


    public MavenPluginClassLoaderProvider(File localRepository, RemoteRepository remoteRepository, RemoteRepository... additionalRemoteRepositories) {
        this.localRepository = localRepository;
        this.remoteRepository = remoteRepository;
        this.additionalRemoteRepositories = Arrays.copyOf(additionalRemoteRepositories, additionalRemoteRepositories.length);
    }

    /**
     * Resolves a Maven artifact (specified as <code>groupId:artifactId:version</code>) from the repository and adds
     * a class loader capable of loading the artifact and its transitive dependencies.
     * @param coordinates the maven artifact coordinates
     */
    public void addMavenArtifact(String coordinates) {
        if (registry != null) {
            addClassLoadersForArtifacts(singletonList(coordinates));
        } else {
            startArtifacts.add(coordinates);
        }
    }

    /**
     * Remove a Maven artifact previously added with {@link #addMavenArtifact(String)}.
     * @param coordinates the maven artifact coordinates
     */
    public void removeMavenArtifact(String coordinates) {
        if (registry != null) {
            removeClassLoadersForArtifacts(singletonList(coordinates));
        } else {
            startArtifacts.remove(coordinates);
        }
    }

    public void start(Registry registry, ClassLoader parentClassLoader) {
        this.registry = registry;
        this.parentClassLoader = parentClassLoader;
        if(startArtifacts.size() > 0) {
            addClassLoadersForArtifacts(startArtifacts);
        }
    }

    private synchronized void addClassLoadersForArtifacts(List<String> coordinates) {
        final RepositorySystem rs = getRepositorySystem();
        RepositorySystemSession session = newRepositorySession(rs);


        Map<String, ClassLoader> classLoaders = new LinkedHashMap<String, ClassLoader>();

        for (String gav : coordinates) {

            Dependency dependency = new Dependency(new DefaultArtifact(gav), "compile");
            CollectRequest collectRequest = new CollectRequest();
            collectRequest.setRoot(dependency);
            collectRequest.addRepository(remoteRepository);
            for(RemoteRepository repository : additionalRemoteRepositories) {
                collectRequest.addRepository(repository);
            }

            try {
                DependencyNode node = rs.collectDependencies(session, collectRequest).getRoot();
                rs.resolveDependencies(session, node, null);


                PreorderNodeListGenerator nlg = new PreorderNodeListGenerator();
                node.accept(nlg);

                classLoaders.put(gav, new URLClassLoader(getURLs(nlg.getFiles()), parentClassLoader));
            } catch (DependencyCollectionException e) {
                handleDependencyCollectionException(e);
            } catch (ArtifactResolutionException e) {
                handleArtifactResolutionException(e);
            }
        }
        this.classLoaders.putAll(classLoaders);
        registry.add(classLoaders.values());

    }

    private synchronized void removeClassLoadersForArtifacts(List<String> coordinates) {
        List<ClassLoader> toRemove = new ArrayList<ClassLoader>();
        for (String coordinate : coordinates) {
            if(classLoaders.containsKey(coordinate)) {
                toRemove.add(classLoaders.get(coordinate));
                classLoaders.remove(coordinate);
            }
        }
        if(!toRemove.isEmpty()) {
            registry.remove(toRemove);
        }
    }

    private void handleArtifactResolutionException(ArtifactResolutionException e) {
        throw new RuntimeException(e);
    }

    private void handleDependencyCollectionException(DependencyCollectionException e) {
        throw new RuntimeException(e);
    }

    public List<String> readVersions(String groupId, String artifactId) {

        final RepositorySystem rs = getRepositorySystem();
        RepositorySystemSession session = newRepositorySession(rs);

        final MetadataRequest mrq = new MetadataRequest(new DefaultMetadata(groupId, artifactId, "maven-metadata.xml", Metadata.Nature.RELEASE), remoteRepository, "hm");

        final List<MetadataResult> resultList = rs.resolveMetadata(session, Collections.singleton(mrq));

        for (MetadataResult metadataResult : resultList) {
            try {
                InputStream fis = new FileInputStream(metadataResult.getMetadata().getFile());

                org.apache.maven.artifact.repository.metadata.Metadata m = new MetadataXpp3Reader().read(fis, false);
                return m.getVersioning().getVersions();

            } catch (FileNotFoundException e) {
                throw new RuntimeException(e);
            } catch (XmlPullParserException e) {
                throw new RuntimeException(e);
            } catch (IOException e) {
                throw new RuntimeException(e);
            }
        }
        return null;

    }

    private URL[] getURLs(List<File> files) {
        List<URL> urls = new ArrayList<URL>(files.size());
        for (File file : files) {
            try {
                urls.add(file.toURI().toURL());
            } catch (MalformedURLException e) {
                throw new RuntimeException(e);
            }
        }
        return urls.toArray(new URL[files.size()]);
    }

    private RepositorySystemSession newRepositorySession(RepositorySystem rs) {

        MavenRepositorySystemSession session = new MavenRepositorySystemSession();
        session.setLocalRepositoryManager(rs.newLocalRepositoryManager(new LocalRepository(localRepository)));

        return session;
    }

    private RepositorySystem getRepositorySystem() {
        DefaultServiceLocator locator = new DefaultServiceLocator();

        locator.addService( RepositoryConnectorFactory.class, AsyncRepositoryConnectorFactory.class );
        locator.addService( RepositoryConnectorFactory.class, FileRepositoryConnectorFactory.class );

        locator.setService(VersionResolver.class, DefaultVersionResolver.class);
        locator.setService(VersionRangeResolver.class, DefaultVersionRangeResolver.class);
        locator.setService(ArtifactDescriptorReader.class, DefaultArtifactDescriptorReader.class);

        return locator.getService( RepositorySystem.class );
    }

    public void stop() {
        registry.remove(classLoaders.values());
        classLoaders.clear();
    }
}
